/*
 * Copyright (c) 1997-1999 Tycho Nightingale.  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The origin of this software must not be misrepresented; you must 
 *    not claim that you wrote the original software.  If you use this 
 *    software in a product, an acknowledgment in the product 
 *    documentation would be appreciated but is not required.
 *
 * 4. Altered source versions must be plainly marked as such, and must
 *    not be misrepresented as being the original software.
 *
 * 5. The name of the author may not be used to endorse or promote 
 *    products derived from this software without specific prior written 
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/*
 * @(#)dcd.c	1.8 03/25/99
 */
/*
 * Include files
 */
#include <sys/scsi/scsi.h>
#include <sys/dkio.h>
#include <sys/vtoc.h>
#include <sys/stat.h>
#include <sys/modctl.h>
#include <sys/file.h>

#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "dcddef.h"

/*
 * Macro definitions
 */

/*
 * Structure definitions
 */
typedef struct {
  kmutex_t lock;
  int id;
} dcd_timeout_t;

typedef struct {
  kmutex_t lock;
  kcondvar_t cv;
  int flag;
} dcd_busy_t;

typedef struct {
  /* devinfo handle */
  dev_info_t	*dip;

  /* layered device related */
  major_t disk_major;
  struct dev_ops *disk_dev_ops;
  struct cb_ops *disk_cb_ops;
  dev_t cache_disk_dev;
  dev_t data_disk_dev;
  
  /* cache related */
  unsigned int log_block_offset;

  /* cache configuration */
  kmutex_t cache_lock;

  int hash_table_size;
  int ram_segment_total_count;
  int disk_segment_total_count;
  
  dcd_line_entry_t **hash_table;

  dcd_disk_segment_t *disk_segment_table;
  dcd_disk_segment_t *disk_segment_free_list;
  dcd_disk_segment_t *disk_segment_dirty_head;
  dcd_disk_segment_t *disk_segment_dirty_tail;
  int disk_segment_dirty_count;

  dcd_ram_segment_t *ram_segment_table;
  dcd_ram_segment_t *ram_segment_free_list;
  dcd_ram_segment_t *ram_segment_dirty_head;
  dcd_ram_segment_t *ram_segment_dirty_tail;

  int current_disk_segment_id;
  int current_ram_segment_id;
  int current_slot;
  
  int current_bg_destage_segment_id;
  int current_fg_destage_segment_id;
  int current_bg_log_write_segment_id;
  int current_fg_log_write_segment_id;

  /* actual cache-data pointers */
  char **ram_segment_data;
  char *destage_segment_data;

  /* bg destage timeout */
  dcd_timeout_t timeout;

  /* busy indicator */
  dcd_busy_t busy;
} dcd_state_t;

/*
 * Forward declarations
 */
static int dcd_identify(dev_info_t *dip);
static int dcd_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static int dcd_attach_layer(dcd_state_t *dcdsp);
static int dcd_attach_cache(dcd_state_t *dcdsp);
static int dcd_attach_timeout(dcd_state_t *dcdsp);
static void dcd_attach_failed(dev_info_t *dip);
static int dcd_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);
static void dcd_detach_layer(dcd_state_t *dcdsp);
static void dcd_detach_cache(dcd_state_t *dcdsp);
static void dcd_detach_timeout(dcd_state_t *dcdsp);
static int dcd_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, 
		      void *arg, void **result);
static int dcd_open(dev_t *devp, int flag, int otyp, cred_t *cred);
static int dcd_open_layer(dcd_state_t *dcdsp, int flag, int otyp, 
			  cred_t *cred);
static int dcd_close(dev_t dev, int flag, int otyp, cred_t *cred);
static void dcd_close_cleanup_segments(dcd_state_t *dcdsp);
static int dcd_close_layer(dcd_state_t *dcdsp, int flag, int otyp,
			   cred_t *cred);
static int dcd_read(dev_t dev, struct uio *uiop, cred_t *credp);
static int dcd_write(dev_t dev, struct uio *uiop, cred_t *credp);
static int dcd_ioctl(dev_t dev, int cmd, intptr_t arg, int mode,
		     cred_t *cred_p, int *rval_p);
static int dcd_print(dev_t dev, char *str);
static int dcd_strategy(struct buf *bp);
static void dcd_read_process(dcd_state_t *dcdsp, buf_t *bp);
static void dcd_do_read_process(dcd_state_t *dcdsp, buf_t *bp);
static int dcd_read_check_entry(dcd_state_t *dcdsp, dcd_line_entry_t *entry,
				int bitmap, int *last_segment_id_p,
				int *last_slot_p);
static int dcd_read_check_frag(dcd_state_t *dcdsp, dcd_line_entry_t *entry,
			       int bitmap, int *last_segment_id_p,
			       int *last_slot_p);
static void dcd_read_from_data_disk(dcd_state_t *dcdsp, buf_t *bp);
static void dcd_read_from_ram(dcd_state_t *dcdsp, buf_t *bp);
static void dcd_read_from_ram_entry(dcd_state_t *dcdsp, 
				    dcd_line_entry_t *entry,
				    buf_t *bp, daddr_t current);
static void dcd_read_from_ram_frag(dcd_state_t *dcdsp, 
				   dcd_line_entry_t *entry, buf_t *bp, 
				   daddr_t current, int bitmap);
static void dcd_read_from_cache_disk(dcd_state_t *dcdsp, buf_t *bp);
static void dcd_do_read_from_cache_disk(dcd_state_t *dcdsp, buf_t *bp,
					int segment_id, int slot);
static void dcd_read_from_combo(dcd_state_t *dcdsp, buf_t *bp);

static void dcd_write_process(dcd_state_t *dcdsp, buf_t *bp);
static void dcd_do_write_process(dcd_state_t *dcdsp, buf_t *bp);
static void dcd_write_memcpy(dcd_state_t *dcdsp, buf_t *bp,
			     daddr_t current);
static void dcd_write_update_entry(dcd_state_t *dcdsp,
				   dcd_line_entry_t *entry,
				   daddr_t block, int bitmap);
static void dcd_write_overwrite_entry(dcd_state_t *dcdsp,
				      dcd_line_entry_t *entry,
				      daddr_t block, int bitmap);
static void dcd_write_new_entry(dcd_state_t *dcdsp, daddr_t block, 
				int bitmap);
static void dcd_do_write_large_process(dcd_state_t *dcdsp, buf_t *bp);
static void dcd_do_write_large(dcd_state_t *dcdsp, buf_t *bp);
static void dcd_write_invalidate_frag(dcd_state_t *dcdsp,
				      dcd_line_entry_t *entry,
				      daddr_t block, int bitmap);
static void dcd_write_invalidate_entry(dcd_state_t *dcdsp,
				       dcd_line_entry_t *entry,
				       daddr_t block, int bitmap);

static void dcd_check_segments(dcd_state_t *dcdsp, buf_t *bp,
			       daddr_t start, daddr_t end);
static void dcd_force_next_segment(dcd_state_t *dcdsp);
static dcd_ram_segment_t *dcd_next_ram_segment(dcd_state_t *dcdsp);
static dcd_disk_segment_t *dcd_next_disk_segment(dcd_state_t *dcdsp);

static void dcd_bg_activity_schedule(dcd_state_t *dcdsp);
static void dcd_bg_activity_deschedule(dcd_state_t *dcdsp);
static void dcd_bg_activity(dcd_state_t *dcdsp);
static void dcd_do_bg_destage(dcd_state_t *dcdsp);
static void dcd_bg_destage_read(dcd_state_t *dcdsp,
				dcd_disk_segment_t *segment);
static void dcd_do_bg_destage_read(dcd_state_t *dcdsp, daddr_t block,
				   unsigned int size, char *addr);
static void dcd_bg_destage_write(dcd_state_t *dcdsp, 
				 dcd_disk_segment_t *segment);
static int dcd_bg_destage_write_frag(dcd_state_t *dcdsp,
				      dcd_disk_segment_t *segment,
				      dcd_line_entry_t *entry,
				      daddr_t block, daddr_t line, int slot);
static void dcd_do_bg_destage_write_frag(dcd_state_t *dcdsp, daddr_t block,
					 unsigned int size, char *addr);
static int dcd_bg_destage_write_entry(dcd_state_t *dcdsp,
				       dcd_disk_segment_t *segment,
				       dcd_line_entry_t *entry,
				       daddr_t block, daddr_t line, int slot);
static void dcd_do_bg_destage_write_entry(dcd_state_t *dcdsp, daddr_t block, 
					  unsigned int size, char *addr);
static void dcd_done_bg_destage_write(dcd_state_t *dcdsp,
				      dcd_disk_segment_t *segment,
				      dcd_line_entry_t *entry, int slot);
static void dcd_bg_destage_done(dcd_state_t *dcdsp, 
				dcd_disk_segment_t *segment);

static void dcd_fg_destage_segment(dcd_state_t *dcdsp, buf_t *bp,
				   dcd_disk_segment_t *segment);
static void dcd_fg_destage_ram_segment(dcd_state_t *dcdsp, buf_t *bp, 
				       dcd_disk_segment_t *disk_segment);
static void dcd_fg_destage_read(dcd_state_t *dcdsp, 
				dcd_disk_segment_t *segment);
static void dcd_do_fg_destage_read(dcd_state_t *dcdsp, daddr_t block,
				   unsigned int size, char *addr);
static void dcd_fg_destage_write(dcd_state_t *dcdsp, buf_t *bp,
				 dcd_disk_segment_t *segment);
static int dcd_fg_destage_write_frag(dcd_state_t *dcdsp, buf_t *bp,
				      dcd_disk_segment_t *segment,
				      dcd_line_entry_t *entry,
				      daddr_t block, daddr_t line, int slot);
static void dcd_do_fg_destage_write_frag(dcd_state_t *dcdsp, daddr_t block,
					 unsigned int size, char *addr);
static int dcd_fg_destage_write_entry(dcd_state_t *dcdsp, buf_t *bp,
				       dcd_disk_segment_t *segment,
				       dcd_line_entry_t *entry,
				       daddr_t block, daddr_t line, int slot);
static void dcd_do_fg_destage_write_entry(dcd_state_t *dcdsp, daddr_t block, 
					  unsigned int size, char *addr);
static void dcd_done_fg_destage_write(dcd_state_t *dcdsp, buf_t *bp,
				      dcd_disk_segment_t *segment,
				      dcd_line_entry_t *entry, int slot);
static void dcd_fg_destage_done(dcd_state_t *dcdsp,  buf_t *bp,
				dcd_disk_segment_t *segment);

static void dcd_bg_log_write(dcd_state_t *dcdsp);
static void dcd_do_bg_log_write(dcd_state_t *dcdsp);
static void dcd_bg_log_write_fix_header(dcd_state_t *dcdsp,
					dcd_ram_segment_t *ram_segment);
static void dcd_bg_log_write_to_cache_disk(dcd_state_t *dcdsp, 
					   dcd_ram_segment_t *ram_segment);
static void dcd_do_bg_log_write_to_cache_disk(dcd_state_t *dcdsp,
					      daddr_t block,
					      unsigned int size,
					      char *addr);
static void dcd_done_bg_log_write(dcd_state_t *dcdsp,
				  dcd_ram_segment_t *ram_segment);

static void dcd_fg_log_write(dcd_state_t *dcdsp, buf_t *bp);
static void dcd_do_fg_log_write(dcd_state_t *dcdsp, buf_t *bp);
static void dcd_fg_log_write_fix_header(dcd_state_t *dcdsp,
					dcd_ram_segment_t *ram_segment);
static void dcd_fg_log_write_to_cache_disk(dcd_state_t *dcdsp, 
					   dcd_ram_segment_t *ram_segment);
static void dcd_do_fg_log_write_to_cache_disk(dcd_state_t *dcdsp,
					      daddr_t block,
					      unsigned int size,
					      char *addr);
static void dcd_done_fg_log_write(dcd_state_t *dcdsp,
				  dcd_ram_segment_t *ram_segment,
				  buf_t *bp);

static dcd_line_entry_t *dcd_line_entry_new(void);
static void dcd_line_entry_free(dcd_line_entry_t *entry);
static dcd_frag_entry_t *dcd_frag_entry_new(void);
static void dcd_frag_entry_free(dcd_frag_entry_t *entry);
static void dcd_log(dev_info_t *dip, const unsigned int level,
                    char *format, ...);

/*
 * Global variables
 */
static void *dcd_state;
static unsigned int Gdebug_level = DEBUG_ERROR | DEBUG_VERBOSE;

/*
 * character/block entry points structure
 */
static struct cb_ops dcd_cb_ops = {
  dcd_open,
  dcd_close,
  dcd_strategy,
  dcd_print,
  nodev,		/* dump */
  dcd_read,
  dcd_write,
  dcd_ioctl,
  nodev,		/* devmap */
  nodev,		/* mmap */
  nodev,		/* segmap */
  nochpoll,		/* poll */
  ddi_prop_op,
  NULL,
  D_NEW | D_MP
};

/*
 * device operations structure
 */
static struct dev_ops dcd_ops = {
  DEVO_REV,
  0,
  dcd_getinfo,
  dcd_identify,
  nulldev,	/* probe */
  dcd_attach,
  dcd_detach,
  nodev,	/* reset */
  &dcd_cb_ops,
  (struct bus_ops *)0
};

extern struct mod_ops mod_driverops;

static struct modldrv modldrv = {
  &mod_driverops,
  "dcd driver v1.0A1",
  &dcd_ops
};

static struct modlinkage modlinkage = {
  MODREV_1,
  {&modldrv, 0}
};

/*
 * Function definitions
 */
int _init(void)
{
  int e;

  if ((e = ddi_soft_state_init(&dcd_state,
			       sizeof(dcd_state_t), 1)) != 0) {
    return (e);
  }

  if ((e = mod_install(&modlinkage)) != 0)  {
    ddi_soft_state_fini(&dcd_state);
  }

  return (e);
}

int _fini(void)
{
  int e;

  if ((e = mod_remove(&modlinkage)) != 0)  {
    return (e);
  }
  ddi_soft_state_fini(&dcd_state);
  return (e);
}

int _info(struct modinfo *modinfop)
{
  return (mod_info(&modlinkage, modinfop));
}

static int dcd_identify(dev_info_t *dip)
{
  if (strcmp(ddi_get_name(dip), "dcd") == 0) {
    return (DDI_IDENTIFIED);
  } else
    return (DDI_NOT_IDENTIFIED);
}

static int dcd_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
  int instance;
  dcd_state_t *dcdsp;

  switch (cmd) {
  case DDI_ATTACH:
    instance = ddi_get_instance(dip);

    if (ddi_soft_state_zalloc(dcd_state, instance) != DDI_SUCCESS) {
      LOG(dip, DEBUG_ERROR, "can not allocate soft state\n");
      return (DDI_FAILURE);
    } else {
      dcdsp = ddi_get_soft_state(dcd_state, instance);
    }

    if ((ddi_create_minor_node(dip, "char", S_IFCHR, instance, NULL, 0)
	 == DDI_FAILURE) ||
	(ddi_create_minor_node(dip, "block", S_IFBLK, instance, NULL, 0)
	 == DDI_FAILURE)) {
      ddi_remove_minor_node(dip, NULL);
      
      dcd_attach_failed(dip);
      return(DDI_FAILURE);
    }

    dcdsp->dip = dip;

    if (dcd_attach_layer(dcdsp) == DDI_FAILURE) {
      dcd_attach_failed(dip);
      return(DDI_FAILURE);
    }

    if (dcd_attach_cache(dcdsp) == DDI_FAILURE) {
      dcd_attach_failed(dip);
      return(DDI_FAILURE);
    }

    if (dcd_attach_timeout(dcdsp) == DDI_FAILURE) {
      dcd_attach_failed(dip);
      return(DDI_FAILURE);
    }

    ddi_report_dev(dip);

    return (DDI_SUCCESS);
  default:
    return (DDI_FAILURE);
  }
}

static int dcd_attach_layer(dcd_state_t *dcdsp)
{
  major_t disk_major;
  struct dev_ops *disk_dev_ops;
  struct cb_ops *disk_cb_ops;
  
  disk_major = ddi_name_to_major("sd");
  disk_dev_ops = ddi_hold_installed_driver(disk_major);
  if (disk_dev_ops == NULL) {
    LOG(dcdsp->dip, DEBUG_ERROR, "failed to find disk dev_ops\n");
    return(DDI_FAILURE);
  }
  disk_cb_ops = disk_dev_ops->devo_cb_ops;

  dcdsp->disk_major = disk_major;
  dcdsp->disk_dev_ops = disk_dev_ops;
  dcdsp->disk_cb_ops = disk_cb_ops;
  
  dcdsp->cache_disk_dev = CACHE_DISK_DEV(disk_major);
  dcdsp->data_disk_dev = DATA_DISK_DEV(disk_major);

  LOG(dcdsp->dip, DEBUG_VERBOSE, "sd driver setup okay\n");
  return(DDI_SUCCESS);
}

static int dcd_attach_cache(dcd_state_t *dcdsp)
{
  int i, j;

  mutex_init(&dcdsp->cache_lock, "", MUTEX_DRIVER, NULL);
  mutex_init(&dcdsp->busy.lock, "", MUTEX_DRIVER, NULL);
  
  cv_init(&dcdsp->busy.cv, "", CV_DRIVER, NULL);

  dcdsp->log_block_offset = LOG_BLOCK_OFFSET;
  
  dcdsp->hash_table_size = HASH_SIZE;
  dcdsp->ram_segment_total_count = RAM_SEGMENTS;
  dcdsp->disk_segment_total_count = DISK_SEGMENTS;

  dcdsp->hash_table = (dcd_line_entry_t **)
    kmem_zalloc(dcdsp->hash_table_size * sizeof(dcd_line_entry_t *),
		KM_SLEEP);
  if (dcdsp->hash_table == NULL) {
    LOG(dcdsp->dip, DEBUG_ERROR, "kmem_zalloc hash_table: failed\n");
    return(DDI_FAILURE);
  }
  
  dcdsp->disk_segment_free_list = NULL;
  dcdsp->disk_segment_dirty_head = NULL;
  dcdsp->disk_segment_dirty_tail = NULL;
  dcdsp->disk_segment_dirty_count = 0;
  
  dcdsp->ram_segment_free_list = NULL;
  dcdsp->ram_segment_dirty_head = NULL;
  dcdsp->ram_segment_dirty_tail = NULL;

  dcdsp->current_disk_segment_id = -1;
  dcdsp->current_ram_segment_id = -1;
  dcdsp->current_slot = -1;
  dcdsp->current_bg_destage_segment_id = -1;
  dcdsp->current_fg_destage_segment_id = -1;
  dcdsp->current_bg_log_write_segment_id = -1;
  dcdsp->current_fg_log_write_segment_id = -1;

  dcdsp->disk_segment_table = (dcd_disk_segment_t *)
    kmem_zalloc(DISK_SEGMENTS * sizeof(dcd_disk_segment_t),
		KM_SLEEP);
  if (dcdsp->disk_segment_table == NULL) {
    LOG(dcdsp->dip, DEBUG_ERROR, "kmem_zalloc disk_segment_table: failed\n");
    return(DDI_FAILURE);
  }
  for (i = (DISK_SEGMENTS - 1); i >= 0; i--) {
    dcdsp->disk_segment_table[i].id = i;
    dcdsp->disk_segment_table[i].ram_id = -1;
    for (j = 0; j < SEGMENT_SLOTS; j++) {
      dcdsp->disk_segment_table[i].slot_lba_table[j] = -1;
    }
    dcdsp->disk_segment_table[i].prev = NULL;
    dcdsp->disk_segment_table[i].next = dcdsp->disk_segment_free_list;
    dcdsp->disk_segment_free_list = &dcdsp->disk_segment_table[i];
  }

  dcdsp->ram_segment_table = (dcd_ram_segment_t *)
    kmem_zalloc(RAM_SEGMENTS * sizeof(dcd_ram_segment_t),
		KM_SLEEP);
  if (dcdsp->ram_segment_table == NULL) {
    LOG(dcdsp->dip, DEBUG_ERROR, "kmem_zalloc ram_segment_table: failed\n");
    return(DDI_FAILURE);
  }
  for (i = (RAM_SEGMENTS - 1); i >= 0; i--) {
    dcdsp->ram_segment_table[i].id = i;
    dcdsp->ram_segment_table[i].disk_id = -1;
    dcdsp->ram_segment_table[i].prev = NULL;
    dcdsp->ram_segment_table[i].next = dcdsp->ram_segment_free_list;
    dcdsp->ram_segment_free_list = &dcdsp->ram_segment_table[i];
  }
  
  dcdsp->ram_segment_data = (char **) 
    kmem_alloc(sizeof(char *) * RAM_SEGMENTS, KM_SLEEP);
  if (dcdsp->ram_segment_data == NULL) {
    LOG(dcdsp->dip, DEBUG_ERROR, "kmem_zalloc ram_segment_data: failed\n");
  }

  for (i = 0; i < RAM_SEGMENTS; i++) {
    dcdsp->ram_segment_data[i] = (char *)
      kmem_alloc(sizeof(char) * DISK_SEG_SIZE, KM_SLEEP);
    if (dcdsp->ram_segment_data[i] == NULL) {
      LOG(dcdsp->dip, DEBUG_ERROR,
	  "kmem_zalloc ram_segment_data[%d]: failed\n", i);
      return(DDI_FAILURE);
    }
  }

  dcdsp->destage_segment_data = (char *)
    kmem_alloc(sizeof(char *) * SEGMENT_SIZE, KM_SLEEP);
  if (dcdsp->destage_segment_data == NULL) {
    LOG(dcdsp->dip, DEBUG_ERROR,
	"kmem_zalloc destage_segment_data: failed\n");
    return(DDI_FAILURE);
  }

  LOG(dcdsp->dip, DEBUG_VERBOSE, "dcd cache setup okay\n");
  LOG(dcdsp->dip, DEBUG_VERBOSE, 
      " %d ram segments (%dkb) %d disk segments (%dkb)\n",
      RAM_SEGMENTS, RAM_SEGMENTS * SEGMENT_SIZE / 1024,
      DISK_SEGMENTS, DISK_SEGMENTS * SEGMENT_SIZE / 1024);
  LOG(dcdsp->dip, DEBUG_VERBOSE, 
      " slot blocks: %d, line blocks: %d, segment slots: %d\n",
      SLOT_BLOCKS, LINE_BLOCKS, SEGMENT_SLOTS);
  LOG(dcdsp->dip, DEBUG_VERBOSE,
      " header blocks: %d, total cache disk blocks: %d\n",
      HEADER_BLOCKS, LOG_BLOCK_OFFSET + (DISK_SEGMENTS * DISK_SEG_BLOCKS));

  return(DDI_SUCCESS);
}

static int dcd_attach_timeout(dcd_state_t *dcdsp)
{
  mutex_init(&dcdsp->timeout.lock, "", MUTEX_DRIVER, NULL);
  dcdsp->timeout.id = 0;

  LOG(dcdsp->dip, DEBUG_VERBOSE, "dcd timeout setup okay\n");
  return(DDI_SUCCESS);
}

static void dcd_attach_failed(dev_info_t *dip)
{
  /*
   * Use our own detach routine to toss
   * away any stuff we allocated above.
   */
  (void) dcd_detach(dip, DDI_DETACH);
}

static int dcd_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
  int instance;
  dcd_state_t *dcdsp;

  switch (cmd) {
  case DDI_DETACH:
    /*
     * Undo what we did in dcd_attach, freeing resources
     * and removing things we installed.  The system
     * framework guarantees we are not active with this devinfo
     * node in any other entry points at this time.
     */
    instance = ddi_get_instance(dip);
    dcdsp = ddi_get_soft_state(dcd_state, instance);
    
    dcd_detach_timeout(dcdsp);
    dcd_detach_cache(dcdsp);
    dcd_detach_layer(dcdsp);
    ddi_remove_minor_node(dip, NULL);
    ddi_soft_state_free(dcd_state, instance);
    return (DDI_SUCCESS);

  default:
    return (DDI_FAILURE);
  }
}

static void dcd_detach_layer(dcd_state_t *dcdsp)
{
  if (dcdsp->disk_dev_ops != NULL) {
    ddi_rele_driver(dcdsp->disk_major);
  } 
}

static void dcd_detach_cache(dcd_state_t *dcdsp)
{
  mutex_destroy(&dcdsp->cache_lock);
  mutex_destroy(&dcdsp->busy.lock);
  cv_destroy(&dcdsp->busy.cv);

  /* FIXME: deallocate the RAM buffer, and all other memory
   * structures. */
}

static void dcd_detach_timeout(dcd_state_t *dcdsp)
{
  /* cancel any pending timeout */
  
  dcd_bg_activity_deschedule(dcdsp);
  
  mutex_destroy(&dcdsp->timeout.lock);
}

/*
 * Device Configuration Routine
 * link instance number (unit) with dev_info structure
 */
static int dcd_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, 
		      void *arg, void **result)
{
  dcd_state_t *dcdsp;
  int instance;
  int error;
  
  switch (infocmd) {
  case DDI_INFO_DEVT2DEVINFO:
    instance = INSTANCE((dev_t)arg);
    if ((dcdsp = ddi_get_soft_state(dcd_state, instance)) != NULL) {
      *result = dcdsp->dip;
      error = DDI_SUCCESS;
    } else
      *result = NULL;
    break;

  case DDI_INFO_DEVT2INSTANCE:
    instance = INSTANCE((dev_t)arg);
    *result = (void *) instance;
    error = DDI_SUCCESS;
    break;

  default:
    error = DDI_FAILURE;
    break;
  }

  return(error);
}

static int dcd_open(dev_t *devp, int flag, int otyp, cred_t *cred)
{
  dcd_state_t *dcdsp;

  LOG(0, DEBUG_ENTRY, "dcd_open - entry\n");  
  dcdsp = ddi_get_soft_state(dcd_state, INSTANCE(*devp));
  if (dcdsp == NULL) {
    LOG(0, DEBUG_ERROR | DEBUG_VERBOSE, "dcd_open - unattached instance %d\n",
	INSTANCE(*devp));
    return (ENXIO);
  }

  if (otyp != OTYP_BLK && otyp != OTYP_CHR) {
    LOG(dcdsp->dip, DEBUG_ERROR | DEBUG_VERBOSE, 
	"dcd_open - unsupported open type %d\n",
	otyp);
    return (EINVAL);
  }

  if (dcd_open_layer(dcdsp, flag, otyp, cred) != DDI_SUCCESS) {
    return(EIO);
  }

  LOG(0, DEBUG_EXIT, "dcd_open - exit\n");  
  return(0);
}

static int dcd_open_layer(dcd_state_t *dcdsp, int flag, int otyp, 
			  cred_t *cred)
{
  int new_otyp;
  int e;
  
  new_otyp = OTYP_BLK;
  e = (*dcdsp->disk_cb_ops->cb_open)(&dcdsp->cache_disk_dev, flag,
				     new_otyp, cred);
  if (e != 0) {
    LOG(dcdsp->dip, DEBUG_ERROR, "failed to open cache disk: %d\n", e);
    return(DDI_FAILURE);
  }

  e = (*dcdsp->disk_cb_ops->cb_open)(&dcdsp->data_disk_dev, flag, 
				     new_otyp, cred);
  if (e != 0) {
    LOG(dcdsp->dip, DEBUG_ERROR, "failed to open data disk: %d\n", e);
    return(DDI_FAILURE);
  }

  return(DDI_SUCCESS);
}

static int dcd_close(dev_t dev, int flag, int otyp, cred_t *cred)
{
  dcd_state_t *dcdsp;

  LOG(0, DEBUG_ENTRY | DEBUG_CLOSE, "dcd_close - entry\n");

  dcdsp = ddi_get_soft_state(dcd_state, INSTANCE(dev));
  if (dcdsp == NULL) {
    LOG(0, DEBUG_ERROR, "dcd_close - unattached instance %d\n",
	INSTANCE(dev));
    return (ENXIO);
  }

  dcd_close_cleanup_segments(dcdsp);

  if (dcd_close_layer(dcdsp, flag, otyp, cred) != DDI_SUCCESS) {
    return(EIO);
  }

  LOG(0, DEBUG_EXIT | DEBUG_CLOSE, "dcd_close - exit\n");  
  return(0);
}

static void dcd_close_cleanup_segments(dcd_state_t *dcdsp)
{
  dcd_ram_segment_t *ram_segment;
  dcd_disk_segment_t *disk_segment;

  mutex_enter(&dcdsp->cache_lock);
  /* clean up ram segments */
  if (dcdsp->current_ram_segment_id != -1) {
    dcd_force_next_segment(dcdsp);
  }
  while (dcdsp->ram_segment_dirty_tail != NULL) {
    dcd_bg_log_write(dcdsp);
  }

  /* clean up disk segments */
  while (dcdsp->disk_segment_dirty_tail != NULL) {
    dcd_do_bg_destage(dcdsp);
  }
  mutex_exit(&dcdsp->cache_lock);
}

static int dcd_close_layer(dcd_state_t *dcdsp, int flag, int otyp, 
			   cred_t *cred)
{
  int new_otyp;
  int e;
	
  new_otyp = OTYP_BLK;
  e = (*dcdsp->disk_cb_ops->cb_close)(dcdsp->cache_disk_dev, flag, 
				      new_otyp, cred);
  if (e != 0) {
    LOG(dcdsp->dip, DEBUG_ERROR, "failed to close cache disk: %d\n", e);
    return(DDI_FAILURE);
  }

  e = (*dcdsp->disk_cb_ops->cb_close)(dcdsp->data_disk_dev, flag, 
				      new_otyp, cred);
  if (e != 0) {
    LOG(dcdsp->dip, DEBUG_ERROR, "failed to close data disk: %d\n", e);
    return(DDI_FAILURE);
  }

  return(DDI_SUCCESS);
}

static int dcd_read(dev_t dev, struct uio *uiop, cred_t *credp)
{
  int instance;
  dcd_state_t *dcdsp;
  int e;

  LOG(0, DEBUG_ENTRY, "dcd_read - entry\n");
  
  instance = INSTANCE(dev);
  dcdsp = ddi_get_soft_state(dcd_state, instance);
  if (dcdsp == NULL) {
    LOG(0, DEBUG_ERROR, "dcd_read -- unattached instance %d\n",
	INSTANCE(dev));
    return (ENXIO);
  }

  e = physio(dcd_strategy, (struct buf *)0, dev, B_READ, minphys, uiop);

  LOG(0, DEBUG_EXIT, "dcd_read - exit\n");  
  return(e);
}

static int dcd_write(dev_t dev, struct uio *uiop, cred_t *credp)
{
  int instance;
  dcd_state_t *dcdsp;
  int e;
 
  LOG(0, DEBUG_ENTRY, "dcd_write - entry\n");

  instance = INSTANCE(dev);
  dcdsp = ddi_get_soft_state(dcd_state, instance);
  if (dcdsp == NULL) {
    LOG(0, DEBUG_ERROR, "dcd_write -- unattached instance %d\n",
	INSTANCE(dev));
    return (ENXIO);
  }

  e = physio(dcd_strategy, (struct buf *)0, dev, B_WRITE, minphys, uiop);

  LOG(0, DEBUG_EXIT, "dcd_write - exit\n");  
  return(e);
}

static int dcd_ioctl(dev_t dev, int cmd, intptr_t arg, int mode,
		     cred_t *cred_p, int *rval_p)
{
  LOG(0, DEBUG_ENTRY, "dcd_ioctl - entry\n");

  LOG(0, DEBUG_EXIT, "dcd_ioctl - exit\n");  
  return(0);
}

static int dcd_print(dev_t dev, char *str)
{
  int instance = INSTANCE(dev);
  dcd_state_t *dcdsp = ddi_get_soft_state(dcd_state, instance);

  cmn_err(CE_WARN, "%s%d: %s\n", ddi_get_name(dcdsp->dip), instance, str);
  return (0);
}

static int dcd_strategy(struct buf *bp)
{
  int instance;
  dcd_state_t *dcdsp;

  LOG(0, DEBUG_ENTRY, "dcd_strategy - entry\n");

  instance = INSTANCE(bp->b_edev);
  dcdsp = ddi_get_soft_state(dcd_state, instance);
  if (dcdsp == NULL) {
    LOG(0, DEBUG_ERROR, "dcd_strategy -- unattached instance %d\n",
	INSTANCE(bp->b_edev));
    bp->b_resid = bp->b_bcount;
    bioerror(bp, ENXIO);
    biodone(bp);
    return (0);
  }

  dcd_bg_activity_deschedule(dcdsp);

  bp_mapin(bp);

  mutex_enter(&dcdsp->busy.lock);
  dcdsp->busy.flag++;
  mutex_exit(&dcdsp->busy.lock);

  if (bp->b_flags & B_READ) {
    dcd_read_process(dcdsp, bp);   
  } else {
    dcd_write_process(dcdsp, bp); 
  }
  
  mutex_enter(&dcdsp->busy.lock);
  dcdsp->busy.flag--;
  mutex_exit(&dcdsp->busy.lock);

  dcd_bg_activity_schedule(dcdsp);

  LOG(0, DEBUG_EXIT, "dcd_strategy - exit\n");
  return(0);
}

static void dcd_read_process(dcd_state_t *dcdsp, buf_t *bp)
{
  daddr_t start, end;

  start = bp->b_blkno;
  end = start + btod(bp->b_bcount);

  /* sanity check the request */
  if (start % SLOT_BLOCKS || end % SLOT_BLOCKS) {
    LOG(0, CE_PANIC, "read unaligned request\n");
  }

  mutex_enter(&dcdsp->cache_lock);

  dcd_do_read_process(dcdsp, bp);
}

static void dcd_do_read_process(dcd_state_t *dcdsp, buf_t *bp)
{
  daddr_t start, end;
  daddr_t current, next;
  int bitmap, bit;
  dcd_line_entry_t *entry;
  int found_flag;
  int last_segment_id, last_slot;
  int entry_count, frag_count;

  found_flag = 0;
  last_segment_id = -1;
  last_slot = -1;

  start = bp->b_blkno;
  end = start + btod(bp->b_bcount);

  entry_count = frag_count = 0;

  for (current = (start & LINE_MASK);
       current < end;
       current += LINE_BLOCKS) { 
    next = current + LINE_BLOCKS;
    bitmap = FRAG_MASK;
    
    /* check for head fragments */
    if (current < start) {
      for (bit = 0;
	   bit < (start - current)/SLOT_BLOCKS;
	   bit++)
	bitmap &= ~(1 << bit);
    }

    /* check for tail fragments */
    if (end < next) {
      for (bit = (end % LINE_BLOCKS)/SLOT_BLOCKS;
	   bit < (end % LINE_BLOCKS)/SLOT_BLOCKS
	     + (next - end)/SLOT_BLOCKS;
	   bit++)
	bitmap &= ~(1 << bit);
    }

    if (bitmap != FRAG_MASK) {
      frag_count += 1;
    } else {
      entry_count += 1;
    }

    /* search hash table */
    for (entry = dcdsp->hash_table[current % dcdsp->hash_table_size];
	 entry != NULL;
	 entry = entry->hash_next) {
      if (entry->block == current) {
	break;
      }
    }
    
    if (entry != NULL) {
      if (entry->frag != NULL) {
	found_flag |= dcd_read_check_frag(dcdsp, entry, bitmap, 
					  &last_segment_id, &last_slot);
      } else {
	found_flag |= dcd_read_check_entry(dcdsp, entry, bitmap,
					   &last_segment_id, &last_slot);
      }
    } else {
      found_flag |= READ_IN_DATA_DISK;
    }
  }

  if ((found_flag | READ_IN_DATA_DISK) == READ_IN_DATA_DISK) {
    /* data is just in data disk */
    dcd_read_from_data_disk(dcdsp, bp);
  } else if (((found_flag & ~READ_IS_NOT_CONTIG) | READ_IN_RAM)
	     == READ_IN_RAM) {
    /* contiguous or not, data is in ram */
    dcd_read_from_ram(dcdsp, bp);
  } else if (((found_flag | READ_IN_CACHE_DISK) == READ_IN_CACHE_DISK)
	     && ((found_flag | READ_IS_NOT_CONTIG) != READ_IS_NOT_CONTIG)) {
    /* data is just in the cache disk and is contiguous */
    dcd_read_from_cache_disk(dcdsp, bp); 
  } else {
    /* data is in more than one location */
    dcd_read_from_combo(dcdsp, bp);
  }
}

static int dcd_read_check_entry(dcd_state_t *dcdsp, dcd_line_entry_t *entry,
				int bitmap, int *last_segment_id_p,
				int *last_slot_p)
{
  dcd_disk_segment_t *segment;
  int segment_id, last_segment_id, slot, last_slot;
  int found_flag;

  last_segment_id = *last_segment_id_p;
  last_slot = *last_slot_p;
  segment_id = entry->segment;
  slot = entry->slot;
  segment = &dcdsp->disk_segment_table[segment_id];
  
  if (segment->ram_id == -1) {
    found_flag = READ_IN_CACHE_DISK;
  } else {
    found_flag = READ_IN_RAM;
  }
  
  if ((last_segment_id != -1 && last_segment_id != segment_id)
      || (last_slot != -1 && (last_slot + 1 != slot))) {
    found_flag |= READ_IS_NOT_CONTIG;
  }

  *last_segment_id_p = segment_id;
  *last_slot_p = slot + SLOTS_PER_LINE - 1;
      
  return(found_flag);
}

static int dcd_read_check_frag(dcd_state_t *dcdsp, dcd_line_entry_t *entry,
			       int bitmap, int *last_segment_id_p,
			       int *last_slot_p)
{
  dcd_disk_segment_t *segment;
  int segment_id, last_segment_id, slot, last_slot;
  int found_flag;
  int i;

  last_segment_id = *last_segment_id_p;
  last_slot = *last_slot_p;

  found_flag = 0;

  for (i = 0; i < SLOTS_PER_LINE; i++) {
    if (bitmap & (1 << i)) {
      if (entry->frag->bitmap & (1 << i)) {
	segment_id = entry->frag->lba_segment_list[i];
	slot = entry->frag->lba_slot_list[i];
	segment = &dcdsp->disk_segment_table[segment_id];
      
	if (segment->ram_id == -1) {
	  found_flag |= READ_IN_CACHE_DISK;
	} else {
	  found_flag |= READ_IN_RAM;
	}

	if ((last_segment_id != -1 && last_segment_id != segment_id)
	    || (last_slot != -1 && (last_slot + 1 != slot))) {
	  found_flag |= READ_IS_NOT_CONTIG;
	}
      
	last_slot = slot;
      } else {
	found_flag |= READ_IN_DATA_DISK;
      }
    }
  }
  
  *last_segment_id_p = segment_id;
  *last_slot_p = slot;

  return(found_flag);
}

static void dcd_read_from_data_disk(dcd_state_t *dcdsp, buf_t *orig_bp)
{
  struct buf *bp = getrbuf(KM_SLEEP);

  bp_mapin(orig_bp);

  bp->b_edev = dcdsp->data_disk_dev; 
  bp->b_flags = B_BUSY | (orig_bp->b_flags & B_READ);
  bp->b_bcount = orig_bp->b_bcount;
  bp->b_blkno = orig_bp->b_blkno;
  bp->b_un.b_addr = orig_bp->b_un.b_addr;

  (*dcdsp->disk_cb_ops->cb_strategy)(bp);

  biowait(bp);

  if (bp->b_resid || geterror(bp)) {
    LOG(0, DEBUG_ERROR, "read data from data disk error: %d resid: %d\n", 
	geterror(bp), bp->b_resid);
    LOG(0, DEBUG_ERROR, " block: %ld count: %d\n", bp->b_blkno, bp->b_bcount);
  }

  freerbuf(bp);

  biodone(orig_bp);

  mutex_exit(&dcdsp->cache_lock);
}

static void dcd_read_from_ram(dcd_state_t *dcdsp, buf_t *bp)
{
  daddr_t start, end;
  daddr_t current, next;
  int bitmap, bit;
  dcd_line_entry_t *entry;

  start = bp->b_blkno;
  end = start + btod(bp->b_bcount);

  for (current = (start & LINE_MASK);
       current < end;
       current += LINE_BLOCKS) { 
    next = current + LINE_BLOCKS;
    bitmap = FRAG_MASK;
    
    /* check for head fragments */
    if (current < start) {
      for (bit = 0;
	   bit < (start - current)/SLOT_BLOCKS;
	   bit++)
	bitmap &= ~(1 << bit);
    }

    /* check for tail fragments */
    if (end < next) {
      for (bit = (end % LINE_BLOCKS)/SLOT_BLOCKS;
	   bit < (end % LINE_BLOCKS)/SLOT_BLOCKS
	     + (next - end)/SLOT_BLOCKS;
	   bit++)
	bitmap &= ~(1 << bit);
    }

    /* search hash table */
    for (entry = dcdsp->hash_table[current % dcdsp->hash_table_size];
	 entry != NULL;
	 entry = entry->hash_next) {
      if (entry->block == current) {
	break;
      }
    }

    if (entry->frag != NULL) {
      dcd_read_from_ram_frag(dcdsp, entry, bp, current, bitmap);
    } else {
      dcd_read_from_ram_entry(dcdsp, entry, bp, current);
    }
  }

  biodone(bp);

  mutex_exit(&dcdsp->cache_lock);  
}

static void dcd_read_from_ram_entry(dcd_state_t *dcdsp, 
				    dcd_line_entry_t *entry,
				    buf_t *bp, daddr_t current)
{
  daddr_t start, end, next;
  int src_offset, dst_offset, length;
  dcd_disk_segment_t *disk_segment;
  int disk_segment_id;
  dcd_ram_segment_t *ram_segment;
  int ram_segment_id;
  int  slot;
  char *dst;
  char *src;
  int size;

  start = bp->b_blkno;
  end = start + btod(bp->b_bcount);
  next = current + LINE_BLOCKS;

  if (start < current) {
    src_offset = 0;
    dst_offset = current - start;
    
    if (end < next) {
      length = end - current;
    } else {
      length = next - current;
    }
  } else {
    src_offset = start - current;
    dst_offset = 0;
    
    if (end < next) {
      length = end - start;
    } else {
      length = next - start;
    }
  } 

  disk_segment_id = entry->segment;
  disk_segment = &dcdsp->disk_segment_table[disk_segment_id];
  ram_segment_id = disk_segment->ram_id;
  ram_segment = &dcdsp->ram_segment_table[ram_segment_id];
  slot = entry->slot;

  src = dcdsp->ram_segment_data[ram_segment_id];
  src += HEADER_SIZE;
  src += slot * SLOT_SIZE + src_offset * DEV_BSIZE;
  dst = bp->b_un.b_addr;
  dst += dst_offset * DEV_BSIZE;
  size = length * DEV_BSIZE;

  bcopy(src, dst, size);
}

static void dcd_read_from_ram_frag(dcd_state_t *dcdsp, 
				   dcd_line_entry_t *entry, buf_t *bp, 
				   daddr_t current, int bitmap)
{
  daddr_t start;
  int offset;
  dcd_disk_segment_t *disk_segment;
  int disk_segment_id;
  dcd_ram_segment_t *ram_segment;
  int ram_segment_id;
  int  slot;
  int i;
  char *dst;
  char *src;
  int size;

  start = bp->b_blkno;
  
  for (i = 0; i < SLOTS_PER_LINE; i++) {
    if (bitmap & (1 << i)) {
      disk_segment_id = entry->frag->lba_segment_list[i];
      disk_segment = &dcdsp->disk_segment_table[disk_segment_id];
      ram_segment_id = disk_segment->ram_id;
      ram_segment = &dcdsp->ram_segment_table[ram_segment_id];
      slot = entry->frag->lba_slot_list[i];

      if (start < (current + i * SLOT_BLOCKS)) {
	offset = (current + i * SLOT_BLOCKS) - start;
      } else {
	offset = 0;
      }

      src = dcdsp->ram_segment_data[ram_segment_id];
      src += HEADER_SIZE;
      src += slot * SLOT_SIZE;
      dst = bp->b_un.b_addr;
      dst += offset * DEV_BSIZE;
      size =  SLOT_SIZE;

      bcopy(src, dst, size);
    }
  }
}

static void dcd_read_from_cache_disk(dcd_state_t *dcdsp, buf_t *bp)
{
  daddr_t start;
  daddr_t current;
  int slot_offset;
  dcd_line_entry_t *entry;
  int segment_id, slot;

  start = bp->b_blkno;
  current = start & LINE_MASK;

  /* search hash table */
  for (entry = dcdsp->hash_table[current % dcdsp->hash_table_size];
       entry != NULL;
       entry = entry->hash_next) {
    if (entry->block == current) {
      break;
    }
  }

  /* find segment and starting slot */
  if (current < start) {
    /* request does not start on a line boundary */
    slot_offset = (start - current) / SLOT_BLOCKS;
    if (entry->frag != NULL) {
      segment_id = entry->frag->lba_segment_list[slot_offset];
      slot = entry->frag->lba_slot_list[slot_offset];
    } else {
      segment_id = entry->segment;
      slot = entry->slot + slot_offset;
    }
  } else {
    /* request does start on a line boundary */
    if (entry->frag != NULL) {
      /* segment has a fragment-entry, check the first entry
       * to see where to start */
      segment_id = entry->frag->lba_segment_list[0];
      slot = entry->frag->lba_slot_list[0];
    } else {
      segment_id = entry->segment;
      slot = entry->slot;
    }
  }

  dcd_do_read_from_cache_disk(dcdsp, bp, segment_id, slot);
}

static void dcd_do_read_from_cache_disk(dcd_state_t *dcdsp, buf_t *orig_bp,
					int segment_id, int slot) 
{
  struct buf *bp = getrbuf(KM_SLEEP);
  daddr_t block;

  block = dcdsp->log_block_offset + (segment_id * DISK_SEG_BLOCKS);
  block += HEADER_BLOCKS + slot * SLOT_BLOCKS;

  bp_mapin(bp);

  bp->b_edev = dcdsp->cache_disk_dev; 
  bp->b_flags = B_BUSY | (orig_bp->b_flags & B_READ);
  bp->b_blkno = block;
  bp->b_bcount = orig_bp->b_bcount;
  bp->b_un.b_addr = orig_bp->b_un.b_addr;

  (*dcdsp->disk_cb_ops->cb_strategy)(bp);

  biowait(bp);

  if (bp->b_resid || geterror(bp)) {
    LOG(0, DEBUG_ERROR, "read data from cache disk error: %d resid: %d\n", 
	geterror(bp), bp->b_resid);
    LOG(0, DEBUG_ERROR, " block: %ld count: %d\n", bp->b_blkno, bp->b_bcount);
  }
  
  freerbuf(bp);

  biodone(orig_bp);

  mutex_exit(&dcdsp->cache_lock);
}

static void dcd_read_from_combo(dcd_state_t *dcdsp, buf_t *bp)
{
  daddr_t start, end;
  daddr_t current;
  dcd_line_entry_t *entry;
  int segment_id;
  dcd_disk_segment_t *segment;
  int i;

  start = bp->b_blkno;
  end = start + btod(bp->b_bcount);

  for (current = (start & LINE_MASK);
       current < end;
       current += LINE_BLOCKS) { 
    /* search hash table */
    for (entry = dcdsp->hash_table[current % dcdsp->hash_table_size];
	 entry != NULL;
	 entry = entry->hash_next) {
      if (entry->block == current) {
	break;
      }
    }

    if (entry != NULL) {
      break;
    }
  }

  if (entry->frag != NULL) {
    /* search for the first entry */
    for (i = 0; i < SLOTS_PER_LINE; i++) {
      if (entry->frag->bitmap & (1 << i)) {
	segment_id = entry->frag->lba_segment_list[i];
	segment = &dcdsp->disk_segment_table[segment_id];
	break;
      }
    }
  } else {
    segment_id = entry->segment;
    segment = &dcdsp->disk_segment_table[segment_id];
  }

  if (segment->ram_id != -1) {
    /* segment is in ram */
    if (segment_id == dcdsp->current_disk_segment_id) {
      /* we are about to destage the current segment */
      dcdsp->current_ram_segment_id = -1;
      dcdsp->current_disk_segment_id = -1;
    }
    
    dcd_fg_destage_ram_segment(dcdsp, bp, segment);
  } else {
    /* segment is on disk */
    dcd_fg_destage_segment(dcdsp, bp, segment);
  }
}

static void dcd_write_process(dcd_state_t *dcdsp, buf_t *bp)
{
  daddr_t start, end;

  start = bp->b_blkno;
  end = start + btod(bp->b_bcount);

  /* sanity check the request */
  if (start % SLOT_BLOCKS || end % SLOT_BLOCKS) {
    LOG(0, CE_PANIC, "unaligned write request\n");
  }

  mutex_enter(&dcdsp->cache_lock);

  if ((end - start) > LARGE_REQUEST_BLOCKS) {
    dcd_do_write_large_process(dcdsp, bp); 
  } else {
    dcd_do_write_process(dcdsp, bp);
  }

}

static void dcd_do_write_process(dcd_state_t *dcdsp, buf_t *bp)
{
  daddr_t start, end;
  daddr_t current, next;
  int bitmap, bit;
  dcd_line_entry_t *entry;

  start = bp->b_blkno;
  end = start + btod(bp->b_bcount);

  dcd_check_segments(dcdsp, bp, start, end);
  
  if (dcdsp->current_ram_segment_id == -1 
      || dcdsp->current_disk_segment_id == -1) {
    dcd_fg_log_write(dcdsp, bp);
    dcd_check_segments(dcdsp, bp, start, end);
  }

  for (current = (start & LINE_MASK);
       current < end;
       current += LINE_BLOCKS) { 
    next = current + LINE_BLOCKS;
    bitmap = FRAG_MASK;
    
    /* check for head fragments */
    if (current < start) {
      for (bit = 0;
	   bit < (start - current)/SLOT_BLOCKS;
	   bit++)
	bitmap &= ~(1 << bit);
    }

    /* check for tail fragments */
    if (end < next) {
      for (bit = (end % LINE_BLOCKS)/SLOT_BLOCKS;
	   bit < (end % LINE_BLOCKS)/SLOT_BLOCKS
	     + (next - end)/SLOT_BLOCKS;
	   bit++)
	bitmap &= ~(1 << bit);
    }
    dcd_write_memcpy(dcdsp, bp, current);

    /* search hash table */
    for (entry = dcdsp->hash_table[current % dcdsp->hash_table_size];
	 entry != NULL;
	 entry = entry->hash_next) {
      if (entry->block == current) {
	break;
      }
    }
    
    if (entry != NULL) {
      /* found something in the table, update or overwrite the entry */
      if (bitmap != FRAG_MASK) {
	/* current request is fragment */
	dcd_write_update_entry(dcdsp, entry, current, bitmap);	
      } else {
	/* current request is not a fragment */
	dcd_write_overwrite_entry(dcdsp, entry, current, bitmap);
      }
    } else {
      /* found nothing in the table, create a new entry */
      dcd_write_new_entry(dcdsp, current, bitmap);
    }
  }  

  biodone(bp);

#if 0
  if (dcdsp->current_slot > SLOT_HIGH_WATERMARK) {
    /* add segment to dirty list */
    dcd_force_next_segment(dcdsp);
  }
#endif
  if (dcdsp->ram_segment_dirty_tail != NULL) {
    /* write the segment now */
    dcd_bg_log_write(dcdsp);
  }

  if (dcdsp->disk_segment_dirty_count > 
      ((dcdsp->disk_segment_total_count >> 1)
       + (dcdsp->disk_segment_total_count >> 2))) {
    dcd_do_bg_destage(dcdsp);
  }

  mutex_exit(&dcdsp->cache_lock);
}

static void dcd_write_memcpy(dcd_state_t *dcdsp, buf_t *bp,
			     daddr_t current)
{
  daddr_t start, end, next;
  int offset, length;
  char *dst;
  char *src;
  int size;
  
  start = bp->b_blkno;
  end = start + btod(bp->b_bcount);
  next = current + LINE_BLOCKS;

  if (start < current) {
    offset = current - start;
    
    if (end < next) {
      length = end - current;
    } else {
      length = next - current;
    }
  } else {
    offset = 0;
    
    if (end < next) {
      length = end - start;
    } else {
      length = next - start;
    }
  } 

  src = bp->b_un.b_addr;
  src += offset * DEV_BSIZE;
  dst = dcdsp->ram_segment_data[dcdsp->current_ram_segment_id];
  dst += HEADER_SIZE;
  dst += dcdsp->current_slot * SLOT_SIZE;
  size = length * DEV_BSIZE;

  bcopy(src, dst, size);
}

static void dcd_write_update_entry(dcd_state_t *dcdsp,
				   dcd_line_entry_t *entry,
				   daddr_t block, int bitmap)
{
  dcd_disk_segment_t *segment, *old_segment;
  int segment_id, slot, old_segment_id, old_slot;
  int i;

  segment_id = dcdsp->current_disk_segment_id;
  segment = &dcdsp->disk_segment_table[segment_id];

  if (entry->frag == NULL) {
    /* allocate a setup a new fragment */
    entry->frag = dcd_frag_entry_new();
    entry->frag->bitmap = FRAG_MASK;
    for (i = 0; i < SLOTS_PER_LINE; i++) { 
      entry->frag->lba_segment_list[i] = entry->segment;
      entry->frag->lba_slot_list[i] = entry->slot + i;
    } 
    entry->segment = -1;
    entry->slot = -1;
  }
  /* merge the new and old fragments */
  for (i = 0; i < SLOTS_PER_LINE; i++) {
    if (bitmap & (1 << i)) {
      slot = dcdsp->current_slot;

      if (entry->frag->bitmap & (1 << i)) {
	old_segment_id = entry->frag->lba_segment_list[i];
	old_segment = &dcdsp->disk_segment_table[old_segment_id];
	old_slot = entry->frag->lba_slot_list[i];

	old_segment->slot_lba_table[old_slot] = -1;
	old_segment->valid_count -= 1;
      } else {
	entry->frag->bitmap |= (1 << i);
      }
      entry->frag->lba_segment_list[i] = segment_id;
      entry->frag->lba_slot_list[i] = slot;

      segment->slot_lba_table[slot] = block + i * SLOT_BLOCKS; 

      segment->valid_count += 1;

      dcdsp->current_slot += 1;
    }
  }
}

static void dcd_write_overwrite_entry(dcd_state_t *dcdsp,
				      dcd_line_entry_t *entry,
				      daddr_t block, int bitmap)
{
  dcd_disk_segment_t *segment, *old_segment;
  int segment_id, slot, old_segment_id, old_slot;
  int i;

  segment_id = dcdsp->current_disk_segment_id;
  segment = &dcdsp->disk_segment_table[segment_id];

  /* new entry is not a fragment, invalidate any previous entry
   * including all fragments */
  if (entry->frag != NULL) {
    for (i = 0; i < SLOTS_PER_LINE; i++) {
      if (entry->frag->bitmap & (1 << i)) {
	old_segment_id = entry->frag->lba_segment_list[i];
	old_slot = entry->frag->lba_slot_list[i];
	old_segment = &dcdsp->disk_segment_table[old_segment_id];
	  
	old_segment->slot_lba_table[old_slot] = -1;
	old_segment->valid_count -= 1;
      }
    }
    dcd_frag_entry_free(entry->frag);
    entry->frag = NULL;
  } else {
    old_segment_id = entry->segment;
    old_slot = entry->slot;
    old_segment = &dcdsp->disk_segment_table[old_segment_id];
    for (i = 0; i < SLOTS_PER_LINE; i++) {
      old_segment->slot_lba_table[old_slot + i] = -1;
      old_segment->valid_count -= 1;
    }
  }

  slot = dcdsp->current_slot;

  entry->segment = segment_id;
  entry->slot = slot;

  for (i = 0; i < SLOTS_PER_LINE; i++) {
    segment->slot_lba_table[slot+i] = block + i*SLOT_BLOCKS;
  }

  dcdsp->current_slot += SLOTS_PER_LINE;

  segment->valid_count += SLOTS_PER_LINE;
}

static void dcd_write_new_entry(dcd_state_t *dcdsp, daddr_t block, 
				int bitmap)
{
  dcd_line_entry_t *entry;
  dcd_disk_segment_t *segment;
  int segment_id, slot;
  int i;

  segment_id = dcdsp->current_disk_segment_id;  
  segment = &dcdsp->disk_segment_table[segment_id];

  entry = dcd_line_entry_new();
  entry->block = block;

  if (bitmap != FRAG_MASK){
    /* new entry is a fragment */
    entry->frag = dcd_frag_entry_new();
    entry->frag->bitmap = bitmap;

    for (i = 0; i < SLOTS_PER_LINE; i++) {
      if (bitmap & (1 << i)) {
	slot = dcdsp->current_slot;

	entry->frag->lba_segment_list[i] = segment_id;
	entry->frag->lba_slot_list[i] = slot;

	segment->slot_lba_table[slot] = block + i*SLOT_BLOCKS;

	dcdsp->current_slot += 1;
	segment->valid_count += 1;
      }
    }
  } else {
    /* new entry is not a fragment */
    slot = dcdsp->current_slot;

    entry->segment = segment_id;
    entry->slot = slot;

    for (i = 0; i < SLOTS_PER_LINE; i++) {
      segment->slot_lba_table[slot+i] = block + i*SLOT_BLOCKS;
    }

    dcdsp->current_slot += SLOTS_PER_LINE;
    segment->valid_count += SLOTS_PER_LINE;
  } 

  entry->key = block % dcdsp->hash_table_size;

  /* insert into hash table */
  entry->hash_next = dcdsp->hash_table[entry->key];
  if (entry->hash_next != NULL) {
    entry->hash_next->hash_prev = entry;
  }
  dcdsp->hash_table[entry->key] = entry;
}

static void dcd_do_write_large_process(dcd_state_t *dcdsp, buf_t *bp)
{
  daddr_t start, end;
  daddr_t current, next;
  int bitmap, bit;
  dcd_line_entry_t *entry;

  start = bp->b_blkno;
  end = start + btod(bp->b_bcount);

  for (current = (start & LINE_MASK);
       current < end;
       current += LINE_BLOCKS) { 
    next = current + LINE_BLOCKS;
    bitmap = FRAG_MASK;
    
    /* check for head fragments */
    if (current < start) {
      for (bit = 0;
	   bit < (start - current)/SLOT_BLOCKS;
	   bit++)
	bitmap &= ~(1 << bit);
    }

    /* check for tail fragments */
    if (end < next) {
      for (bit = (end % LINE_BLOCKS)/SLOT_BLOCKS;
	   bit < (end % LINE_BLOCKS)/SLOT_BLOCKS
	     + (next - end)/SLOT_BLOCKS;
	   bit++)
	bitmap &= ~(1 << bit);
    }

    /* search hash table */
    for (entry = dcdsp->hash_table[current % dcdsp->hash_table_size];
	 entry != NULL;
	 entry = entry->hash_next) {
      if (entry->block == current) {
	break;
      }
    }
    
    if (entry != NULL) {
      /* found something in the table, update or overwrite the entry */
      if (bitmap != FRAG_MASK) {
	/* current request is fragment */
	dcd_write_invalidate_frag(dcdsp, entry, current, bitmap);
      } else {
	/* current request is not a fragment */
	dcd_write_invalidate_entry(dcdsp, entry, current, bitmap);
      }
    } 
  }

  dcd_do_write_large(dcdsp, bp);
}

static void dcd_do_write_large(dcd_state_t *dcdsp, buf_t *orig_bp)
{
  struct buf *bp = getrbuf(KM_SLEEP);
  
  bp_mapin(bp);

  bp->b_edev = dcdsp->data_disk_dev; 
  bp->b_flags = B_BUSY | (orig_bp->b_flags & B_READ);
  bp->b_bcount = orig_bp->b_bcount;
  bp->b_blkno = orig_bp->b_blkno;
  bp->b_un.b_addr = orig_bp->b_un.b_addr;

  (*dcdsp->disk_cb_ops->cb_strategy)(bp);
  
  biowait(bp);

  if (bp->b_resid || geterror(bp)) {
    LOG(0, DEBUG_ERROR, "write data to data disk error: %d resid: %d\n", 
	geterror(bp), bp->b_resid);
  }
  
  freerbuf(bp);
 
  biodone(orig_bp);

  mutex_exit(&dcdsp->cache_lock);  
}

static void dcd_write_invalidate_frag(dcd_state_t *dcdsp,
				      dcd_line_entry_t *entry,
				      daddr_t block, int bitmap)
{
  dcd_disk_segment_t *old_segment;
  int old_segment_id, old_slot;
  int i;

  if (entry->frag == NULL) {
    /* allocate a setup a new fragment */
    entry->frag = dcd_frag_entry_new();
    entry->frag->bitmap = FRAG_MASK;
    for (i = 0; i < SLOTS_PER_LINE; i++) { 
      entry->frag->lba_segment_list[i] = entry->segment;
      entry->frag->lba_slot_list[i] = entry->slot + i;
    } 
    entry->segment = -1;
    entry->slot = -1;
  }
  /* merge the new and old fragments */
  for (i = 0; i < SLOTS_PER_LINE; i++) {
    if (bitmap & (1 << i)) {
      if (entry->frag->bitmap & (1 << i)) {
	old_segment_id = entry->frag->lba_segment_list[i];
	old_segment = &dcdsp->disk_segment_table[old_segment_id];
	old_slot = entry->frag->lba_slot_list[i];

	old_segment->slot_lba_table[old_slot] = -1;
	old_segment->valid_count -= 1;

	entry->frag->lba_segment_list[i] = -1;
	entry->frag->lba_slot_list[i] = -1;
	entry->frag->bitmap &= ~(1 << i);
      }
    }
  }

  /* check to see if the entry should be free from the hash table */
  if (entry->frag->bitmap == 0) {
    if (entry->frag != NULL) {
      dcd_frag_entry_free(entry->frag);
    }

    /* remove from the hash table */
    if (entry->hash_prev != NULL) {
      entry->hash_prev->hash_next = entry->hash_next;
    } else {
      dcdsp->hash_table[entry->key] = entry->hash_next;
    }
    
    if (entry->hash_next != NULL) {
      entry->hash_next->hash_prev = entry->hash_prev;
    }

    dcd_line_entry_free(entry);
  }

}

static void dcd_write_invalidate_entry(dcd_state_t *dcdsp,
				       dcd_line_entry_t *entry,
				       daddr_t block, int bitmap)
{
  dcd_disk_segment_t *old_segment;
  int old_segment_id, old_slot;
  int i;

  /* invalidate any previous entry including all fragments */
  if (entry->frag != NULL) {
    for (i = 0; i < SLOTS_PER_LINE; i++) {
      if (entry->frag->bitmap & (1 << i)) {
	old_segment_id = entry->frag->lba_segment_list[i];
	old_slot = entry->frag->lba_slot_list[i];
	old_segment = &dcdsp->disk_segment_table[old_segment_id];
	  
	old_segment->slot_lba_table[old_slot] = -1;
	old_segment->valid_count -= 1;
      }
    }
    dcd_frag_entry_free(entry->frag);
    entry->frag = NULL;
  } else {
    old_segment_id = entry->segment;
    old_slot = entry->slot;
    old_segment = &dcdsp->disk_segment_table[old_segment_id];
    for (i = 0; i < SLOTS_PER_LINE; i++) {
      old_segment->slot_lba_table[old_slot + i] = -1;
      old_segment->valid_count -= 1;
    }
  }

  /* remove from the hash table */
  if (entry->hash_prev != NULL) {
    entry->hash_prev->hash_next = entry->hash_next;
  } else {
    dcdsp->hash_table[entry->key] = entry->hash_next;
  }
  
  if (entry->hash_next != NULL) {
    entry->hash_next->hash_prev = entry->hash_prev;
  }
  
  dcd_line_entry_free(entry);
}

static void dcd_check_segments(dcd_state_t *dcdsp, buf_t *bp,
			       daddr_t start, daddr_t end)
{
  dcd_ram_segment_t *ram_segment;
  dcd_disk_segment_t *disk_segment;
  
  if (dcdsp->current_ram_segment_id != -1
      && dcdsp->current_disk_segment_id != -1
      && (end - start)/SLOT_BLOCKS
      <= (SEGMENT_SLOTS - dcdsp->current_slot)) {
    /* segments are all in order */
    return;
  }
  
  if (dcdsp->current_ram_segment_id != -1
      && dcdsp->current_disk_segment_id != -1
      && (end - start)/SLOT_BLOCKS
      > (SEGMENT_SLOTS - dcdsp->current_slot)) {
    dcd_force_next_segment(dcdsp);
  }
  
  ram_segment = dcd_next_ram_segment(dcdsp);
  if (ram_segment == NULL) {
    /* must wait */
    return;
  }

  disk_segment = dcd_next_disk_segment(dcdsp);
  ASSERT(disk_segment);

  /* link ram and disk segments and set current */
  ram_segment->disk_id = disk_segment->id;
  disk_segment->ram_id = ram_segment->id;

  dcdsp->current_ram_segment_id = ram_segment->id;
  dcdsp->current_disk_segment_id = disk_segment->id;
  dcdsp->current_slot = 0;
}

static void dcd_force_next_segment(dcd_state_t *dcdsp)
{
  dcd_ram_segment_t *ram_segment;
  int ram_segment_id;
  
  ram_segment_id = dcdsp->current_ram_segment_id;
  ram_segment = &dcdsp->ram_segment_table[ram_segment_id];
  
  ram_segment->prev = NULL; 
  ram_segment->next = NULL; 

  /* add current ram segment to the dirty fifo */
  if (dcdsp->ram_segment_dirty_tail == NULL) {
    dcdsp->ram_segment_dirty_tail = ram_segment;
  } else {
    ram_segment->next = dcdsp->ram_segment_dirty_head;
    dcdsp->ram_segment_dirty_head->prev = ram_segment;
  }
  dcdsp->ram_segment_dirty_head = ram_segment;

  dcdsp->current_ram_segment_id = -1;
  dcdsp->current_disk_segment_id = -1;
}

static dcd_ram_segment_t *dcd_next_ram_segment(dcd_state_t *dcdsp)
{
  dcd_ram_segment_t *segment;

  if (dcdsp->ram_segment_free_list != NULL) {
    /* take a segment off the free list */
    segment = dcdsp->ram_segment_free_list;
    dcdsp->ram_segment_free_list = segment->next;

    segment->next = NULL;
    segment->prev = NULL;
  } else {
    segment = NULL;
  }

  return(segment);
}

static dcd_disk_segment_t *dcd_next_disk_segment(dcd_state_t *dcdsp)
{
  dcd_disk_segment_t *segment;
  
  if (dcdsp->disk_segment_free_list != NULL) {
    /* take a segment off the free list */
    segment = dcdsp->disk_segment_free_list;
    dcdsp->disk_segment_free_list = segment->next;
    
    segment->next = NULL;
    segment->prev = NULL;
  } else {
    segment = NULL;
  }

  return(segment);
}

static void dcd_bg_activity_schedule(dcd_state_t *dcdsp)
{
  int id;
  long ticks;


  mutex_enter(&dcdsp->timeout.lock);
  id = dcdsp->timeout.id; 
  mutex_exit(&dcdsp->timeout.lock);
  
  if (id != 0) {
    (void) untimeout(id);
  }    
  
  mutex_enter(&dcdsp->cache_lock);
  if (dcdsp->disk_segment_dirty_count > 
      (dcdsp->disk_segment_total_count >> 1)) {
    ticks = DESTAGE_NORMAL_TIMEOUT;
  } else {
    ticks = DESTAGE_SLOW_TIMEOUT;
  }
  mutex_exit(&dcdsp->cache_lock);

  id = timeout(dcd_bg_activity, (caddr_t) dcdsp, ticks);

  mutex_enter(&dcdsp->timeout.lock);
  dcdsp->timeout.id = id;
  mutex_exit(&dcdsp->timeout.lock);
}

static void dcd_bg_activity_deschedule(dcd_state_t *dcdsp)
{
  int id;
  
  mutex_enter(&dcdsp->timeout.lock);  
  id = dcdsp->timeout.id;
  mutex_exit(&dcdsp->timeout.lock);
  
  if (id != 0) {
    (void) untimeout(id);
    
    mutex_enter(&dcdsp->timeout.lock);  
    dcdsp->timeout.id = 0;
    mutex_exit(&dcdsp->timeout.lock);
  }
}

static void dcd_bg_activity(dcd_state_t *dcdsp)
{
  mutex_enter(&dcdsp->timeout.lock);  
  dcdsp->timeout.id = 0;
  mutex_exit(&dcdsp->timeout.lock);

  mutex_enter(&dcdsp->busy.lock);
  if (dcdsp->busy.flag) {
    mutex_exit(&dcdsp->busy.lock);
    return;
  }
  mutex_exit(&dcdsp->busy.lock);

  mutex_enter(&dcdsp->cache_lock);
  if (dcdsp->current_ram_segment_id != -1) { 
    dcd_force_next_segment(dcdsp);    
  } 
  /* write ram segments to cache disk */
  while (dcdsp->ram_segment_dirty_tail != NULL) { 
    dcd_bg_log_write(dcdsp); 
    mutex_enter(&dcdsp->busy.lock); 
    if (dcdsp->busy.flag) { 
      mutex_exit(&dcdsp->busy.lock); 
      mutex_exit(&dcdsp->cache_lock);  
      return; 
    } 
    mutex_exit(&dcdsp->busy.lock); 
  } 
  /* destage cache disk */
  while (dcdsp->disk_segment_dirty_tail != NULL) {
    dcd_do_bg_destage(dcdsp);
    mutex_enter(&dcdsp->busy.lock);
    if (dcdsp->busy.flag) {
      mutex_exit(&dcdsp->busy.lock);
      mutex_exit(&dcdsp->cache_lock); 
      return;
    }
    mutex_exit(&dcdsp->busy.lock);
  } 

  mutex_exit(&dcdsp->cache_lock); 
}

static void dcd_do_bg_destage(dcd_state_t *dcdsp)
{
  dcd_disk_segment_t *segment;
  
  /* take a segment from the end of the fifo */
  segment = dcdsp->disk_segment_dirty_tail;
  if (dcdsp->disk_segment_dirty_tail->prev == NULL) {
    dcdsp->disk_segment_dirty_head = NULL;
  } else {
    dcdsp->disk_segment_dirty_tail->prev->next = NULL;
  }
  dcdsp->disk_segment_dirty_tail = dcdsp->disk_segment_dirty_tail->prev;
  dcdsp->disk_segment_dirty_count -= 1;
  
  segment->prev = NULL;
  segment->next = NULL;

  dcdsp->current_bg_destage_segment_id = segment->id;

  dcd_bg_destage_read(dcdsp, segment); 
  dcd_bg_destage_write(dcdsp, segment);

  dcd_bg_destage_done(dcdsp, segment);
}

static void dcd_bg_destage_read(dcd_state_t *dcdsp,
				dcd_disk_segment_t *segment)
{
  daddr_t block;
  unsigned int size;
  char *addr;
  
  block = dcdsp->log_block_offset + (segment->id * DISK_SEG_BLOCKS);
  block += HEADER_BLOCKS;
  size = SEGMENT_SIZE;
  addr = dcdsp->destage_segment_data;

  dcd_do_bg_destage_read(dcdsp, block, size, addr);
}

static void dcd_do_bg_destage_read(dcd_state_t *dcdsp, daddr_t block,
				   unsigned int size, char *addr)
{
  buf_t *bp;
  
  bp = getrbuf(KM_SLEEP);
  bp_mapin(bp);

  bp->b_flags = (B_BUSY | B_READ);
  bp->b_bcount = size;
  bp->b_un.b_addr = addr;
  bp->b_blkno = block;
  bp->b_edev = dcdsp->cache_disk_dev;

  (*dcdsp->disk_cb_ops->cb_strategy)(bp);
  
  biowait(bp);

  if (bp->b_resid || geterror(bp)) {
    LOG(0, DEBUG_ERROR, "bg destage read error: %d resid: %d\n", 
	geterror(bp), bp->b_resid);
  }

  freerbuf(bp);
}

static void dcd_bg_destage_write(dcd_state_t *dcdsp, 
				 dcd_disk_segment_t *segment)
{
  dcd_line_entry_t *entry;
  daddr_t block, line;
  int slot, slot_incr;

  slot = 0;
  block = -1;

  while (slot < SEGMENT_SLOTS) {
    /* skip all invalid slots */
    for (; slot < SEGMENT_SLOTS; slot++) {
      block = segment->slot_lba_table[slot];
      if (block != -1) {
	break;
      }
    }

    if (block != -1) {
      line = block & LINE_MASK;

      /* search hash table */
      for (entry = dcdsp->hash_table[line % dcdsp->hash_table_size];
	   entry != NULL;
	   entry = entry->hash_next) {
	if (entry->block == line) {
	  break;
	}
      }

      ASSERT(entry);

      if (entry->frag != NULL) {
	slot_incr = dcd_bg_destage_write_frag(dcdsp, segment, entry,
					      block, line, slot); 
      } else {
	slot_incr = dcd_bg_destage_write_entry(dcdsp, segment, entry,
					       block, line, slot); 
      }

      slot += slot_incr;
    }
  }
}

static int dcd_bg_destage_write_frag(dcd_state_t *dcdsp,
				      dcd_disk_segment_t *segment,
				      dcd_line_entry_t *entry,
				      daddr_t block, daddr_t line, int slot)
{
  int bit;
  unsigned int size;
  char *addr;
  
  bit = (block - line) / SLOT_BLOCKS;
  entry->frag->bitmap &= ~(1 << bit);
  entry->frag->lba_segment_list[bit] = -1;
  entry->frag->lba_slot_list[bit] = -1;
  
  size = dtob(SLOT_BLOCKS);
  addr = dcdsp->destage_segment_data;
  addr += slot * SLOT_SIZE;

  dcd_do_bg_destage_write_frag(dcdsp, block, size, addr);
  
  dcd_done_bg_destage_write(dcdsp, segment, entry, slot + 1);
  
  return(1);
}

static void dcd_do_bg_destage_write_frag(dcd_state_t *dcdsp, daddr_t block,
					 unsigned int size, char *addr)
{
  buf_t *bp;
  
  bp = getrbuf(KM_SLEEP);
  bp_mapin(bp);

  bp->b_flags = (B_BUSY | B_WRITE);
  bp->b_bcount = size;
  bp->b_un.b_addr = addr;
  bp->b_blkno = block;
  bp->b_edev = dcdsp->data_disk_dev;

  (*dcdsp->disk_cb_ops->cb_strategy)(bp);
  
  biowait(bp);

  if (bp->b_resid || geterror(bp)) {
    LOG(0, DEBUG_ERROR, "bg destage write frag error: %d resid: %d\n", 
	geterror(bp), bp->b_resid);
  }

  freerbuf(bp);
}

static int dcd_bg_destage_write_entry(dcd_state_t *dcdsp,
				       dcd_disk_segment_t *segment,
				       dcd_line_entry_t *entry,
				       daddr_t block, daddr_t line, int slot)
{
  unsigned int size;
  char *addr;

  size = dtob(LINE_BLOCKS);
  addr = dcdsp->destage_segment_data;
  addr += slot * SLOT_SIZE;

  dcd_do_bg_destage_write_entry(dcdsp, block, size, addr);

  dcd_done_bg_destage_write(dcdsp, segment, entry, slot + SLOTS_PER_LINE);

  return(SLOTS_PER_LINE);
}

static void dcd_do_bg_destage_write_entry(dcd_state_t *dcdsp, daddr_t block, 
					  unsigned int size, char *addr)
{
  buf_t *bp;
  
  bp = getrbuf(KM_SLEEP);
  bp_mapin(bp);

  bp->b_flags = (B_BUSY | B_WRITE);
  bp->b_bcount = size;
  bp->b_un.b_addr = addr;
  bp->b_blkno = block;
  bp->b_edev = dcdsp->data_disk_dev;

  (*dcdsp->disk_cb_ops->cb_strategy)(bp);
  
  biowait(bp);

  if (bp->b_resid || geterror(bp)) {
    LOG(0, DEBUG_ERROR, "bg destage write entry error: %d resid: %d\n", 
	geterror(bp), bp->b_resid);
  }

  freerbuf(bp);
}

static void dcd_done_bg_destage_write(dcd_state_t *dcdsp,
				      dcd_disk_segment_t *segment,
				      dcd_line_entry_t *entry, int slot)
{
  if (entry->frag == NULL || entry->frag->bitmap == 0) {
    if (entry->frag != NULL) {
      dcd_frag_entry_free(entry->frag);
    }

    /* remove from the hash table */
    if (entry->hash_prev != NULL) {
      entry->hash_prev->hash_next = entry->hash_next;
    } else {
      dcdsp->hash_table[entry->key] = entry->hash_next;
    }
    
    if (entry->hash_next != NULL) {
      entry->hash_next->hash_prev = entry->hash_prev;
    }

    dcd_line_entry_free(entry);
  }
}

static void dcd_bg_destage_done(dcd_state_t *dcdsp, 
				dcd_disk_segment_t *segment)
{
  int i;

  dcdsp->current_bg_destage_segment_id = -1;

  /* cleanup segment */
  for (i = 0; i < SEGMENT_SLOTS; i++) {
    segment->slot_lba_table[i] = -1;
  }
  segment->valid_count = 0;

  /* put segment on the free list */
  segment->next = dcdsp->disk_segment_free_list;
  dcdsp->disk_segment_free_list = segment;
}

static void dcd_fg_destage_segment(dcd_state_t *dcdsp, buf_t *bp,
				   dcd_disk_segment_t *segment)
{
  /* remove segment from dirty list */
  if (segment->prev == NULL) {
    dcdsp->disk_segment_dirty_head = segment->next;
  } else {
    segment->prev->next = segment->next;
  }
  
  if (segment->next == NULL) {
    dcdsp->disk_segment_dirty_tail = segment->prev;
  } else {
    segment->next->prev = segment->prev;
  }
  dcdsp->disk_segment_dirty_count -= 1;

  segment->next = NULL;
  segment->prev = NULL;

  dcdsp->current_fg_destage_segment_id = segment->id;

  dcd_fg_destage_read(dcdsp, segment);
  dcd_fg_destage_write(dcdsp, bp, segment);
  
  dcd_fg_destage_done(dcdsp, bp, segment);

  /* continue read */
  dcd_do_read_process(dcdsp, bp);
}

static void dcd_fg_destage_ram_segment(dcd_state_t *dcdsp, buf_t *bp, 
				       dcd_disk_segment_t *disk_segment)
{
  dcd_ram_segment_t *ram_segment, *dirty_list_ram_segment;
  int ram_segment_id;
  char *src, *dst;
  int size;

  dcdsp->current_fg_destage_segment_id = disk_segment->id;

  ram_segment_id = disk_segment->ram_id;
  ram_segment = &dcdsp->ram_segment_table[ram_segment_id];

  /* copy the ram buffer into the destage buffer */
  src = dcdsp->ram_segment_data[ram_segment_id];
  src += HEADER_SIZE;
  dst = dcdsp->destage_segment_data;
  size = SEGMENT_SIZE;

  bcopy(src, dst, size);

  /* check to see if the ram segment is on the dirty list */
  for (dirty_list_ram_segment = dcdsp->ram_segment_dirty_head;
       dirty_list_ram_segment != NULL;
       dirty_list_ram_segment = dirty_list_ram_segment->next) {
    if (dirty_list_ram_segment == ram_segment) {
      LOG(0, DEBUG_VERBOSE, " ram segment on dirty list!\n");
      break;
    }
  }
  /* if it is on the dirty list, remove it */
  if (dirty_list_ram_segment != NULL) {
    if (dirty_list_ram_segment->prev == NULL) {
      dcdsp->ram_segment_dirty_head = dirty_list_ram_segment->next;
    } else {
      dirty_list_ram_segment->prev->next = dirty_list_ram_segment->next;
    }
  
    if (dirty_list_ram_segment->next == NULL) {
      dcdsp->ram_segment_dirty_tail = dirty_list_ram_segment->prev;
    } else {
      dirty_list_ram_segment->next->prev = dirty_list_ram_segment->prev;
    }
  }
  
  ram_segment->disk_id = -1;
  disk_segment->ram_id = -1;

  ram_segment->next = NULL;
  ram_segment->prev = NULL;

  /* put ram segment on the free list */
  ram_segment->next = dcdsp->ram_segment_free_list;
  dcdsp->ram_segment_free_list = ram_segment;

  dcd_fg_destage_write(dcdsp, bp, disk_segment);
  dcd_fg_destage_done(dcdsp, bp, disk_segment);

  /* continue read */
  dcd_do_read_process(dcdsp, bp);
}

static void dcd_fg_destage_read(dcd_state_t *dcdsp, 
				dcd_disk_segment_t *segment)
{
  daddr_t block;
  unsigned int size;
  char *addr;

  block = dcdsp->log_block_offset + (segment->id * DISK_SEG_BLOCKS);
  block += HEADER_BLOCKS;
  size = SEGMENT_SIZE;
  addr = dcdsp->destage_segment_data;

  dcd_do_fg_destage_read(dcdsp, block, size, addr);
}

static void dcd_do_fg_destage_read(dcd_state_t *dcdsp, daddr_t block,
				   unsigned int size, char *addr)
{
  buf_t *bp;
  
  bp = getrbuf(KM_SLEEP);
  bp_mapin(bp);

  bp->b_flags = (B_BUSY | B_READ);
  bp->b_bcount = size;
  bp->b_un.b_addr = addr;
  bp->b_blkno = block;
  bp->b_edev = dcdsp->cache_disk_dev;

  (*dcdsp->disk_cb_ops->cb_strategy)(bp);
  
  biowait(bp);

  if (bp->b_resid || geterror(bp)) {
    LOG(0, DEBUG_ERROR, "fg destage read error: %d resid: %d\n", 
	geterror(bp), bp->b_resid);
  }

  freerbuf(bp);
}

static void dcd_fg_destage_write(dcd_state_t *dcdsp, buf_t *bp,
				 dcd_disk_segment_t *segment)
{
  dcd_line_entry_t *entry;
  daddr_t line, block;
  int slot, slot_incr;
   
  slot = 0;
  block = -1;

  while (slot < SEGMENT_SLOTS) {
    /* skip all invalid slots */
    for (; slot < SEGMENT_SLOTS; slot++) {
      block = segment->slot_lba_table[slot];
      if (block != -1) {
	break;
      }
    }

    if (block != -1) {
      line = block & LINE_MASK;

      /* search hash table */
      for (entry = dcdsp->hash_table[line % dcdsp->hash_table_size];
	   entry != NULL;
	   entry = entry->hash_next) {
	if (entry->block == line) {
	  break;
	}
      }

      ASSERT(entry);

      if (entry->frag != NULL) {
	slot_incr = dcd_fg_destage_write_frag(dcdsp, bp, segment, entry,
					      block, line, slot);
      } else {
	slot_incr = dcd_fg_destage_write_entry(dcdsp, bp, segment, entry,
					       block, line, slot);
      }

      slot += slot_incr;
    }
  }
}

static int dcd_fg_destage_write_frag(dcd_state_t *dcdsp, buf_t *bp,
				      dcd_disk_segment_t *segment,
				      dcd_line_entry_t *entry,
				      daddr_t block, daddr_t line, int slot)
{
  int bit;
  unsigned int size;
  char *addr;
  
  bit = (block - line) / SLOT_BLOCKS;
  entry->frag->bitmap &= ~(1 << bit);
  entry->frag->lba_segment_list[bit] = -1;
  entry->frag->lba_slot_list[bit] = -1;

  size = dtob(SLOT_BLOCKS);
  addr = dcdsp->destage_segment_data;
  addr += slot * SLOT_SIZE;

  dcd_do_fg_destage_write_frag(dcdsp, block, size, addr);
  
  dcd_done_fg_destage_write(dcdsp, bp, segment, entry, slot + 1);

  return(1);
}

static void dcd_do_fg_destage_write_frag(dcd_state_t *dcdsp, daddr_t block,
					 unsigned int size, char *addr)
{
  buf_t *bp;
  
  bp = getrbuf(KM_SLEEP);
  bp_mapin(bp);

  bp->b_flags = (B_BUSY | B_WRITE);
  bp->b_bcount = size;
  bp->b_un.b_addr = addr;
  bp->b_blkno = block;
  bp->b_edev = dcdsp->data_disk_dev;

  (*dcdsp->disk_cb_ops->cb_strategy)(bp);
  
  biowait(bp);

  if (bp->b_resid || geterror(bp)) {
    LOG(0, DEBUG_ERROR, "fg destage write frag error: %d resid: %d\n", 
	geterror(bp), bp->b_resid);
  }

  freerbuf(bp);
}

static int dcd_fg_destage_write_entry(dcd_state_t *dcdsp, buf_t *bp,
				       dcd_disk_segment_t *segment,
				       dcd_line_entry_t *entry,
				       daddr_t block, daddr_t line, int slot)
{
  unsigned int size;
  char *addr;

  size = dtob(LINE_BLOCKS);
  addr = dcdsp->destage_segment_data;
  addr += slot * SLOT_SIZE;

  dcd_do_fg_destage_write_entry(dcdsp, block, size, addr);

  dcd_done_fg_destage_write(dcdsp, bp, segment, entry, slot + SLOTS_PER_LINE);

  return(SLOTS_PER_LINE);
}

static void dcd_do_fg_destage_write_entry(dcd_state_t *dcdsp, daddr_t block, 
					  unsigned int size, char *addr)
{
  buf_t *bp;
  
  bp = getrbuf(KM_SLEEP);
  bp_mapin(bp);

  bp->b_flags = (B_BUSY | B_WRITE);
  bp->b_bcount = size;
  bp->b_un.b_addr = addr;
  bp->b_blkno = block;
  bp->b_edev = dcdsp->data_disk_dev;

  (*dcdsp->disk_cb_ops->cb_strategy)(bp);
  
  biowait(bp);

  if (bp->b_resid || geterror(bp)) {
    LOG(0, DEBUG_ERROR, "fg destage write entry error: %d resid: %d\n", 
	geterror(bp), bp->b_resid);
  }

  freerbuf(bp);
}

static void dcd_done_fg_destage_write(dcd_state_t *dcdsp, buf_t *bp,
				      dcd_disk_segment_t *segment,
				      dcd_line_entry_t *entry, int slot)
{
  if (entry->frag == NULL || entry->frag->bitmap == 0) {
    if (entry->frag != NULL) {
      dcd_frag_entry_free(entry->frag);
    }

    /* remove from the hash table */
    if (entry->hash_prev != NULL) {
      entry->hash_prev->hash_next = entry->hash_next;
    } else {
      dcdsp->hash_table[entry->key] = entry->hash_next;
    }
    
    if (entry->hash_next != NULL) {
      entry->hash_next->hash_prev = entry->hash_prev;
    }

    dcd_line_entry_free(entry);
  }
}

static void dcd_fg_destage_done(dcd_state_t *dcdsp, buf_t *bp,
				dcd_disk_segment_t *segment)
{
  int i;

  /*   printf("fg destage done: %d\n", segment->id); */

  dcdsp->current_fg_destage_segment_id = -1;

  /* cleanup segment */
  for (i = 0; i < SEGMENT_SLOTS; i++) {
    segment->slot_lba_table[i] = -1;
  }
  segment->valid_count = 0;

  /* put disk segment on the free list */
  segment->next = dcdsp->disk_segment_free_list;
  dcdsp->disk_segment_free_list = segment;
}

static void dcd_bg_log_write(dcd_state_t *dcdsp)
{
  ASSERT(mutex_owned(&dcdsp->cache_lock)); 

  if (dcdsp->ram_segment_dirty_tail == NULL) {
    /* nothing to do */
    return;
  }

  dcd_do_bg_log_write(dcdsp);
}

static void dcd_do_bg_log_write(dcd_state_t  *dcdsp)
{
  dcd_ram_segment_t *ram_segment;
  
  /* take a segment from the end of the fifo */
  ram_segment = dcdsp->ram_segment_dirty_tail;
  if (dcdsp->ram_segment_dirty_tail->prev == NULL) {
    dcdsp->ram_segment_dirty_head = NULL;
  } else {
    dcdsp->ram_segment_dirty_tail->prev->next = NULL;
  }
  dcdsp->ram_segment_dirty_tail = dcdsp->ram_segment_dirty_tail->prev;

  ram_segment->prev = NULL;
  ram_segment->next = NULL;

  dcdsp->current_bg_log_write_segment_id = ram_segment->id;

  dcd_bg_log_write_fix_header(dcdsp, ram_segment);
  dcd_bg_log_write_to_cache_disk(dcdsp, ram_segment);
  dcd_done_bg_log_write(dcdsp, ram_segment);
}

static void dcd_bg_log_write_fix_header(dcd_state_t *dcdsp,
					dcd_ram_segment_t *ram_segment)
{
  int ram_segment_id;
  int disk_segment_id;
  dcd_disk_segment_t *disk_segment;
  dcd_segment_header_t *segment_header;
  char *addr;

  ram_segment_id = ram_segment->id;
  disk_segment_id = ram_segment->disk_id;
  disk_segment = &dcdsp->disk_segment_table[disk_segment_id];
  
  addr = dcdsp->ram_segment_data[ram_segment_id];
  segment_header = (dcd_segment_header_t *)addr;
  
  segment_header->id = disk_segment_id;
  bcopy(disk_segment->slot_lba_table,
	segment_header->slot_lba_table,
	sizeof(disk_segment->slot_lba_table));
}

static void dcd_bg_log_write_to_cache_disk(dcd_state_t *dcdsp, 
					   dcd_ram_segment_t *ram_segment)
{
  daddr_t block;
  unsigned int size;
  int ram_segment_id;
  int disk_segment_id;
  char *addr;

  ram_segment_id = ram_segment->id;
  disk_segment_id = ram_segment->disk_id;

  block = dcdsp->log_block_offset + (disk_segment_id * DISK_SEG_BLOCKS);
  size = DISK_SEG_SIZE;
  addr = dcdsp->ram_segment_data[ram_segment_id];

  dcd_do_bg_log_write_to_cache_disk(dcdsp, block, size, addr);
}

static void dcd_do_bg_log_write_to_cache_disk(dcd_state_t *dcdsp,
					      daddr_t block,
					      unsigned int size,
					      char *addr)
{
  buf_t *bp;
  
  bp = getrbuf(KM_SLEEP);
  bp_mapin(bp);

  bp->b_flags = (B_BUSY | B_WRITE);
  bp->b_bcount = size;
  bp->b_un.b_addr = addr;
  bp->b_blkno = block;
  bp->b_edev = dcdsp->cache_disk_dev;

  (*dcdsp->disk_cb_ops->cb_strategy)(bp);
  
  biowait(bp);

  if (bp->b_resid || geterror(bp)) {
    LOG(0, DEBUG_ERROR, "bg log write error: %d resid: %d b: %d s: %u\n", 
	geterror(bp), bp->b_resid, bp->b_blkno, bp->b_bcount);
  }

  freerbuf(bp);
}

static void dcd_done_bg_log_write(dcd_state_t *dcdsp,
				  dcd_ram_segment_t *ram_segment)
{
  dcd_disk_segment_t *disk_segment;
  int disk_segment_id;
    
  disk_segment_id = ram_segment->disk_id;
  disk_segment = &dcdsp->disk_segment_table[disk_segment_id];

  ram_segment->disk_id = -1;
  disk_segment->ram_id = -1;

  ram_segment->next = NULL;
  ram_segment->prev = NULL;

  /* insert ram segment into free list */
  ram_segment->next = dcdsp->ram_segment_free_list;
  dcdsp->ram_segment_free_list = ram_segment;

  ASSERT(dcdsp->current_ram_segment_id != ram_segment->id);

  disk_segment->next = NULL;
  disk_segment->prev = NULL;

  /* insert disk segment into dirty fifo */  
  if (dcdsp->disk_segment_dirty_tail == NULL) {
    dcdsp->disk_segment_dirty_tail = disk_segment;
  } else {
    disk_segment->next = dcdsp->disk_segment_dirty_head;
    dcdsp->disk_segment_dirty_head->prev = disk_segment;
  }
  dcdsp->disk_segment_dirty_head = disk_segment;
  dcdsp->disk_segment_dirty_count += 1;
  
  dcdsp->current_bg_log_write_segment_id = -1;
}

static void dcd_fg_log_write(dcd_state_t *dcdsp, buf_t *bp)
{
  ASSERT(mutex_owned(&dcdsp->cache_lock)); 
  
  ASSERT(dcdsp->ram_segment_dirty_tail != NULL);
  
  dcd_do_fg_log_write(dcdsp, bp);
}

static void dcd_do_fg_log_write(dcd_state_t *dcdsp, buf_t *bp)
{
  dcd_ram_segment_t *ram_segment;
  
  /* take a segment from the end of the fifo */
  ram_segment = dcdsp->ram_segment_dirty_tail;
  if (dcdsp->ram_segment_dirty_tail->prev == NULL) {
    dcdsp->ram_segment_dirty_head = NULL;
  } else {
    dcdsp->ram_segment_dirty_tail->prev->next = NULL;
  }
  dcdsp->ram_segment_dirty_tail = dcdsp->ram_segment_dirty_tail->prev;

  ram_segment->prev = NULL;
  ram_segment->next = NULL;

  dcdsp->current_fg_log_write_segment_id = ram_segment->id;

  dcd_fg_log_write_fix_header(dcdsp, ram_segment);
  dcd_fg_log_write_to_cache_disk(dcdsp, ram_segment);
  dcd_done_fg_log_write(dcdsp, ram_segment, bp);
}

static void dcd_fg_log_write_fix_header(dcd_state_t *dcdsp,
					dcd_ram_segment_t *ram_segment)
{
  int ram_segment_id;
  int disk_segment_id;
  dcd_disk_segment_t *disk_segment;
  dcd_segment_header_t *segment_header;
  char *addr;

  ram_segment_id = ram_segment->id;
  disk_segment_id = ram_segment->disk_id;
  disk_segment = &dcdsp->disk_segment_table[disk_segment_id];
  
  addr = dcdsp->ram_segment_data[ram_segment_id];
  segment_header = (dcd_segment_header_t *)addr;
  
  segment_header->id = disk_segment_id;
  bcopy(disk_segment->slot_lba_table,
	segment_header->slot_lba_table,
	sizeof(disk_segment->slot_lba_table));
}

static void dcd_fg_log_write_to_cache_disk(dcd_state_t *dcdsp, 
					   dcd_ram_segment_t *ram_segment)
{
  daddr_t block;
  unsigned int size;
  int ram_segment_id;
  int disk_segment_id;
  char *addr;

  ram_segment_id = ram_segment->id;
  disk_segment_id = ram_segment->disk_id;

  block = dcdsp->log_block_offset + (disk_segment_id * DISK_SEG_BLOCKS);
  size =  DISK_SEG_SIZE;
  addr = dcdsp->ram_segment_data[ram_segment_id];

  dcd_do_fg_log_write_to_cache_disk(dcdsp, block, size, addr);
}

static void dcd_do_fg_log_write_to_cache_disk(dcd_state_t *dcdsp,
					      daddr_t block,
					      unsigned int size,
					      char *addr)
{
  buf_t *bp;
  
  bp = getrbuf(KM_SLEEP);
  bp_mapin(bp);

  bp->b_flags = (B_BUSY | B_WRITE);
  bp->b_bcount = size;
  bp->b_un.b_addr = addr;
  bp->b_blkno = block;
  bp->b_edev = dcdsp->cache_disk_dev;

  (*dcdsp->disk_cb_ops->cb_strategy)(bp);
  
  biowait(bp);

  if (bp->b_resid || geterror(bp)) {
    LOG(0, DEBUG_ERROR, "fg log write error: %d resid: %d\n", 
	geterror(bp), bp->b_resid);
  }

  freerbuf(bp);
}

static void dcd_done_fg_log_write(dcd_state_t *dcdsp,
				  dcd_ram_segment_t *ram_segment,
				  buf_t *bp)
{
  dcd_disk_segment_t *disk_segment;
  int disk_segment_id;

  disk_segment_id = ram_segment->disk_id;
  disk_segment = &dcdsp->disk_segment_table[disk_segment_id];

  ram_segment->disk_id = -1;
  disk_segment->ram_id = -1;

  ram_segment->next = NULL;
  ram_segment->prev = NULL;

  /* insert ram segment into free list */
  ram_segment->next = dcdsp->ram_segment_free_list;
  dcdsp->ram_segment_free_list = ram_segment;

  ASSERT(dcdsp->current_ram_segment_id != ram_segment->id);

  disk_segment->next = NULL;
  disk_segment->prev = NULL;

  /* insert disk segment into dirty fifo */  
  if (dcdsp->disk_segment_dirty_tail == NULL) {
    dcdsp->disk_segment_dirty_tail = disk_segment;
  } else {
    disk_segment->next = dcdsp->disk_segment_dirty_head;
    dcdsp->disk_segment_dirty_head->prev = disk_segment;
  }
  dcdsp->disk_segment_dirty_head = disk_segment;
  dcdsp->disk_segment_dirty_count += 1;
  
  dcdsp->current_fg_log_write_segment_id = -1;
  
}

static dcd_line_entry_t *dcd_line_entry_new(void)
{
  dcd_line_entry_t *entry;

  entry = (dcd_line_entry_t *) kmem_alloc(sizeof(dcd_line_entry_t), KM_SLEEP);
  if (entry == NULL) {
    LOG(0, DEBUG_ERROR, "kmem_alloc dcd_line_entry_t: failed\n");
  }

  entry->block = -1;
  entry->segment = -1;
  entry->slot = -1;
  entry->frag = NULL;
  entry->key = -1;

  entry->hash_prev = NULL;
  entry->hash_next = NULL;
  
  return(entry);
}
 
static void dcd_line_entry_free(dcd_line_entry_t *entry)
{
  kmem_free(entry, sizeof(dcd_line_entry_t));
}


static dcd_frag_entry_t *dcd_frag_entry_new(void)
{
  dcd_frag_entry_t *entry;
  int i;

  entry = (dcd_frag_entry_t *) kmem_alloc(sizeof(dcd_frag_entry_t), KM_SLEEP);
  if (entry == NULL) {
    LOG(0, DEBUG_ERROR, "kmem_alloc dcd_frag_entry_t: failed\n");
  }

  entry->bitmap = 0;
  for (i = 0; i < SLOTS_PER_LINE; i++) {
    entry->lba_segment_list[i] = -1;
    entry->lba_slot_list[i] = -1;
  }

  return(entry);
}
 
static void dcd_frag_entry_free(dcd_frag_entry_t *entry)
{
  kmem_free(entry, sizeof(dcd_frag_entry_t));
}

static void dcd_log(dev_info_t *dip, const unsigned int level,
                    char *format, ...)
{
  char name[16], buf[256];
  va_list vararg;
  
  if (dip) {
    (void) sprintf(name, "%s%d", ddi_get_name(dip),
                   ddi_get_instance(dip));
  } else {
    (void) sprintf(name, "dcd");
  }
  
  va_start(vararg, format);
  (void) vsprintf(buf, format, vararg);
  va_end(ap);
  
  switch (level) {
  case CE_CONT:
  case CE_NOTE:
  case CE_WARN:
  case CE_PANIC:
    cmn_err(level, "%s:\t%s", name, buf);
    break;
  default:
    if ((Gdebug_level & level)) {
      cmn_err(CE_CONT, "^%s:\t%s", name, buf);
    }
    break;
  }
}
